(* 
Authors: Brody Berg, Max Czapanskiy after Edwards
Date: Fall 2008
*)
open Ast
open String

module NameMap = Map.Make(struct
  type t = string
  let compare x y = Pervasives.compare x y
end)

exception ReturnException of string

let run states =
  (* Put the states into a symbol table *)
  let state_declarations = List.fold_left
      (fun state_map state_declaration -> NameMap.add state_declaration.sname state_declaration state_map)
      NameMap.empty states
  in
  
  (* eval returns a tuple: (result of expression, updated environment) *)
  let rec eval stmt env = match stmt with
     Int(x) -> string_of_int x, env
   | String(x) -> x, env
   | LibCallP(x) ->  
       let value, _ = eval x env in
         print_endline value; value, env
   | LibCallIS(x) -> let value, _ = eval x env in value, env
   | Binop(left, op, right) -> 
          let a, env = eval left env in let x =
          (try int_of_string a with Failure "int_of_string" -> 0)
          in 
          let b, env = eval right env in let y = 
          (try int_of_string b with Failure "int_of_string" -> 0)
          in
          (match op with
            Add -> string_of_int (x + y), env
          | Sub -> string_of_int (x - y), env
          | Mult -> string_of_int (x * y), env
          | Div -> string_of_int (x / y), env
          | Mod -> string_of_int (x mod y), env
          | Assign -> b, NameMap.add 
                (match left with 
                  Id(a) -> a 
                 | _ -> let e = "CABG: Invalid lvalue" in raise (ReturnException e)) b env
          | Less -> string_of_bool (x < y), env
          | Greater -> string_of_bool (x > y), env
          | Equal -> string_of_bool (x == y), env
          | Neq -> string_of_bool (x != y), env )
   | Id(x) -> if NameMap.mem x env 
              then NameMap.find x env, env 
              else x, env
  in  
 
  let rec evalStatementList statementlist env = match statementlist with
    [] -> env
  | s::sl -> let _, result_env = eval s env in evalStatementList sl result_env
  in

  let string_of_statement s = match s with 
      Id(x) -> x 
    | _ -> ""
  in

  (* if the condition is true, iterate over the list of actions, evaluating each one, then go to the target state *)
  let rec follow transitionlist env this = match transitionlist with
    [] -> raise (Failure ("No matching conditions in state " ^ this))
  | t::transitionlist -> let result, _ = eval t.condition env in
             let name = t.target.state in
             if bool_of_string result 
             then let env = evalStatementList t.actions env in 
                  (if compare name "" != 0
                  then 
                    try exec (NameMap.find name state_declarations) 
                             (List.map (fun arg -> let a, _ = eval arg env in a) t.target.args)
                    with Not_found -> raise (Failure ("Did not find state: " ^ name)) )
             else follow transitionlist env this
  and  

  exec s args =
    (* iterate through the formals and add their values to the env *)
    let env = List.fold_left2
      (fun env left right -> let left, _ = eval left env in
                             NameMap.add left right env) 
      NameMap.empty s.formals args in
    (* iterate over statement list, evaluating each line *)
    let env = List.fold_left 
      (fun env stmt -> let _, env = eval stmt env in env) env s.statements; 
    in
    (* iterate over transition list until a valid transition is found *)
    follow s.transitions env s.sname
  in 

  try
    exec (NameMap.find "Start" state_declarations) []
  with Not_found -> raise (Failure ("Did not find the Start state"))
  
